#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2022:
#     Gabes Jean, naparuba@gmail.com
#     Gerhard Lausser, Gerhard.Lausser@consol.de
#     Gregory Starck, g.starck@gmail.com
#     Hartmut Goebel, h.goebel@goebel-consult.de
#
# This file is part of Shinken.
#
# Shinken is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Shinken is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Shinken.  If not, see <http://www.gnu.org/licenses/>.

import imp
import inspect
import json
import os
import sys
import time
from threading import RLock, Thread

from shinken.log import LoggerFactory
from shinken.misc.type_hint import TYPE_CHECKING
from shinken.modules.base_module.basemodule import BaseModule, ModuleState, INSTANCE_INIT_STATE
from shinken.subprocess_helper.after_fork_cleanup import AFTER_FORK_CLEANUP_PART_LOGGER, AFTER_FORK_CLEANUP_METHOD

if TYPE_CHECKING:
    from shinken.misc.type_hint import Any, List, Optional, Union
    from shinken.withworkersandinventorymodule import WithWorkersAndInventoryModule
    from shinken.objects.module import Module as ModuleConfiguration
    from shinken.log import PartLogger
    
    ModuleCode = Any  # Type for module code. Used only for type hint


# We need to manage pre-2.0 module types with _ into the new 2.0 - mode
def uniform_module_type(s):
    return s.replace(u'_', u'-')


class StateCleaner(Thread):
    # Used to cleanup the modules state (self.last_restarts)
    
    CLEAN_STEP = 60  # every 60 seconds
    
    
    def __init__(self, manager):
        # type: (ModulesManager) -> None
        Thread.__init__(self)
        self.setDaemon(True)  # Will be killed at application close
        self.__manager = manager
    
    
    def run(self):
        while True:
            time.sleep(StateCleaner.CLEAN_STEP)
            self.__manager.clean_modules_states()


class ModulesManager(object):
    # This class is use to manage modules and call callback
    
    def __init__(self, modules_type, modules_path, modules, daemon_display_name=u'UNSET', with_thread=True):
        self.logger = LoggerFactory.get_logger(u'MODULES-MANAGER')
        self.logger_module_code_loading = self.logger.get_sub_part(u'CODE-LOADING')
        self.logger_module_instance_creation = self.logger.get_sub_part(u'CREATION')
        self.logger_modules_instance_crash = self.logger.get_sub_part(u'CRASH')
        self.logger_modules_change = self.logger.get_sub_part(u'UPDATE')
        self.logger_modules_stop = self.logger.get_sub_part(u'SHUTDOWN')
        self._modules = None
        self.modules_path = modules_path
        self.modules_type = modules_type
        self.modules = modules
        self.allowed_types = [uniform_module_type(plug.module_type) for plug in modules if hasattr(plug, 'module_type')]
        self.imported_modules = []
        self.modules_assoc = {}
        self.instances = {}
        self.to_restart = {}
        self.last_restarts = {}
        self.last_restarts_lock = RLock()
        self.instances_lock = RLock()
        self.last_restarts_keep = 86400  # lasts 24h (86400 seconds) of restarts
        self.max_queue_size = 0
        self.manager_factory = None
        self.errors = []
        self.daemon_display_name = daemon_display_name
        if with_thread:
            StateCleaner(self).start()
    
    
    @property
    def modules(self):
        return self._modules
    
    
    @modules.setter
    def modules(self, modules):
        self._modules = modules
        self._set_modules_ordering_indexes(modules)
    
    
    def _set_modules_ordering_indexes(self, modules):
        self._idx_modules_order_by_name = {}
        for index, module in enumerate(modules):
            self._idx_modules_order_by_name[module.module_name] = index
    
    
    def load_manager_factory(self, factory):
        self.manager_factory = factory
    
    
    # Set the modules requested for this manager
    def set_modules(self, conf_modules):
        # type: (List) -> None
        self.modules = conf_modules
        self.errors = []
        self.allowed_types = [uniform_module_type(mod.module_type) for mod in conf_modules if hasattr(mod, u'module_type')]
        
        self._display_module_tree(added_modules_names=[module.get_name() for module in self.modules], updated_modules_names=[], deleted_modules_names=[])
    
    
    def _find_module_by_name(self, module_name):
        for module in self.modules:
            if module.get_name() == module_name:
                return module
        return None
    
    
    # When module does changes in a way, we print what we currently have as module
    def _display_module_tree(self, added_modules_names, updated_modules_names, deleted_modules_names):
        # If no modules, don't need to display any thing
        if len(self.modules) == 0 and len(deleted_modules_names) == 0:
            return
        
        # Maybe there was no change in modules
        if len(added_modules_names) == 0 and len(updated_modules_names) == 0 and len(deleted_modules_names) == 0:
            return
        
        # Ok something DID changed
        all_module_names = [module.get_name() for module in self.modules]
        all_module_names.extend(deleted_modules_names)  # all = current have + deleted
        all_module_names = list(set(all_module_names))
        all_module_names.sort()
        self.logger_modules_change.info('Modules did changed:')
        for module_name in all_module_names:
            module = self._find_module_by_name(module_name)
            if module is None:
                what = '-> stopped   ( removed )'
            elif module_name in added_modules_names:
                what = '-> started   ( new )'
            elif module_name in updated_modules_names:
                what = '-> restarted ( configuration change )'
            else:  # untouched
                what = ''
            self.logger_modules_change.info('   - %-20s %s' % (module_name, what))
            if module is None:  # was deleted, don't print sub modules
                continue
            # Maybe there is no sub modules to print
            if len(module.modules) == 0:
                continue
            
            # IMPORTANT: DO NOT SORT sub modules, because for webui for example the
            #            order IS important (for auth modules for example)
            # NOTE: see this is only useful in DEBUG phase
            for sub_module in module.modules:
                self.logger_modules_change.debug('      * %s' % (sub_module.get_name()))
    
    
    def get_module_by_name(self, module_name):
        inst = self.instances.get(module_name, None)
        return inst
    
    
    # We get some new modules, and we must look which one are:
    # * no more there and should be stopped
    # * the new ones, we need to start
    # * the one that did changed (look at hash property) that we need to restart
    def update_modules(self, modules):
        with self.instances_lock:
            changed_to_restart = []
            self.errors = []
            
            # First try to look which one are no more need
            new_modules_name = set([m.get_name() for m in modules])
            old_modules_name = set([m.get_name() for m in self.modules])
            
            old_modules_to_delete_name = old_modules_name - new_modules_name
            new_modules_to_start_name = new_modules_name - old_modules_name
            
            # In the common ones, look at the one that did changed, and so we will have to delete
            module_old_and_new_name = old_modules_name & new_modules_name
            
            for _current_module_name in module_old_and_new_name:
                _logger = self.logger_modules_change.get_sub_part(_current_module_name, register=False)
                old_one = self._find_module_by_name(_current_module_name)
                new_one = next((m for m in modules if m.get_name() == _current_module_name), None)
                
                if old_one is None or new_one is None:
                    _logger.error(u'We cannot update \'%s\' module. Please contact your Support.' % _current_module_name)
                    continue
                # we compare hash to know if some thing did change, if so we will need to kill/restart it
                if old_one.hash != new_one.hash:
                    _logger.info(u'Configuration of the module did change so we restart it.')
                    changed_to_restart.append(new_one)
            
            for _current_module_name in old_modules_to_delete_name:
                _logger = self.logger_modules_change.get_sub_part(_current_module_name, register=False)
                _logger.info(u'Removing the module because it is removed in the configuration.')
                self.__remove_instance(_current_module_name, self.logger_modules_change)
            
            for _current_module_name in new_modules_to_start_name:
                _logger = self.logger_modules_change.get_sub_part(_current_module_name, register=False)
                _logger.info(u'Launched as a NEW module as it was added in configuration.')
                mod_conf = next((m for m in modules if m.get_name() == _current_module_name), None)
                # Got a new module configuration, need to start it
                inst = self.__get_instance_from_modconf(mod_conf)
                if inst is None:
                    continue
                updated_instances = self.instances.copy()
                updated_instances[_current_module_name] = inst
                self.instances = updated_instances
                # Set this module to be restart so it will start
                self.set_to_restart(inst, is_an_error=False)  # here it's not a problem, so don't log it as AT RISK in health check
            
            for new_instance in changed_to_restart:
                _current_module_name = new_instance.get_name()
                _logger = self.logger_modules_change.get_sub_part(_current_module_name, register=False)
                _logger.debug(u'The module did change its configuration. We restart it to take the new configuration.')
                self.__remove_instance(_current_module_name, self.logger_modules_change)
                mod_conf = next((m for m in modules if m.get_name() == _current_module_name), None)
                # Got a new module configuration, need to start it
                inst = self.__get_instance_from_modconf(mod_conf)
                if inst is None:
                    continue
                updated_instances = self.instances.copy()
                updated_instances[_current_module_name] = inst
                self.instances = updated_instances
                if inst.is_external:
                    self.start_external_instance(inst)
                elif inst.is_worker_based:
                    self.start_worker_based_instance(inst)
                else:
                    self.start_internal_instance(inst)
            # save the new modules value
            self.modules = modules
            
            # Now display what did changed
            updated_instances_names = [module.get_name() for module in changed_to_restart]
            self._display_module_tree(added_modules_names=new_modules_to_start_name, updated_modules_names=updated_instances_names, deleted_modules_names=old_modules_to_delete_name)
    
    
    def set_max_queue_size(self, max_queue_size):
        self.max_queue_size = max_queue_size
    
    
    # Import, instantiate & "init" the modules we have been requested
    def load_and_init(self):
        self.load()
        all_was_started = self._get_instances()
        return all_was_started
    
    
    def load(self):
        self._load_python_modules()
    
    
    # We want to know which sys.modules are new after the modules load, but beware of duplicate ones :)
    @staticmethod
    def _get_only_new_sys_modules(sys_modules_names_before, sys_modules_code_before):
        # Look at sys.modules and look which modules are now loaded when they was not before:
        # WARNING: there will be duplicate: entries that match the same lib object
        sys_modules_names_after = set(sys.modules.keys())
        sys_modules_names_added = list(sys_modules_names_after - sys_modules_names_before)
        sys_modules_names_added.sort()
        
        for sys_module_name in list(sys_modules_names_added):  # copy because we will remove elements inside
            sys_module = sys.modules.get(sys_module_name, None)  # maybe a thread did remove it?
            if sys_module in sys_modules_code_before:  # was already loaded in fact
                sys_modules_names_added.remove(sys_module_name)
        return sys_modules_names_added
    
    
    # Try to import the requested modules ; put the imported modules in self.imported_modules.
    def _load_python_modules(self):
        # type: () -> None
        # Was already done, don't load it again
        if self.imported_modules:
            return
        
        # And directories
        modules_dirs = [module_dir_name for module_dir_name in os.listdir(self.modules_path) if os.path.isdir(os.path.join(self.modules_path, module_dir_name))]
        
        # Now we try to load them
        # So first we add their dir into the sys.path
        if self.modules_path not in sys.path:
            sys.path.append(self.modules_path)
        
        before = time.time()
        failed_modules_dir = []
        available_modules_names = []
        # We try to import them, but we keep only the one of
        # our type
        for module_dir_name in modules_dirs:
            _logger = self.logger_module_code_loading.get_sub_part(u'directory=%s' % module_dir_name, register=False)
            module_start = time.time()
            sys_modules_names_before = set(sys.modules.keys())
            sys_modules_code_before = set(sys.modules.values())
            module_info = None
            module_info_path = os.path.join(self.modules_path, module_dir_name, 'module_info.json')
            try:
                if os.path.exists(module_info_path):
                    with open(module_info_path) as json_file:
                        module_info = json.load(json_file)
                        if module_info.get(u'module_types_by_daemons', None):
                            modules_matches = [m for m in module_info[u'module_types_by_daemons'] if m[u'daemon'] == self.modules_type]
                            if not modules_matches:
                                continue
                        elif module_info.get(u'daemons'):
                            if self.modules_type not in module_info[u'daemons']:
                                continue
                        else:
                            _logger.warning(u'Missing \'daemons\' key in the \'module_info.json\' file at the path %s. The code of the module will be loaded.' % module_info_path)
                else:
                    _logger.warning(u'Missing \'module_info.json\' file at the path %s. The code of the module will be loaded.' % module_info_path)
            except ValueError:
                _logger.warning(u'The \'module_info.json\' file at the path %s is malformed. The code of the module will be loaded.' % module_info_path)
            except IOError:
                _logger.warning(u'Permission denied to read file \'module_info.json\' file at the path %s. The code of the module will be loaded.' % module_info_path)
            
            module_dir = os.path.join(self.modules_path, module_dir_name)
            init_path_file = os.path.join(module_dir, '__init__.py')
            
            try:
                _logger.info(u'Starting to load the module code directory %s.' % module_dir)
                
                sys.path.append(module_dir)
                _logger.info(u'Try to import a module code in %s as python module.' % module_dir)
                try:
                    imp.load_module(module_dir_name, *imp.find_module(module_dir_name, [self.modules_path]))
                    _logger.info(u'Success to import the directory %s as a python module.' % module_dir)
                except Exception:
                    _logger.info(u'Failed to import the directory %s as a python module. If it doesn\'t use python module feature, it won\'t encounter loading problems.' % module_dir)
                    pass
                
                # Then we load the module.py inside this directory
                mod_file = os.path.abspath(os.path.join(self.modules_path, module_dir_name, 'module.py'))
                mod_dir = os.path.dirname(mod_file)
                # We add this dir to sys.path so the module can load local files too
                sys.path.append(mod_dir)
                if not os.path.exists(mod_file):
                    mod_file = os.path.abspath(os.path.join(self.modules_path, module_dir_name, 'module.pyc'))
                
                if mod_file.endswith('.py'):
                    # important, equivalent to import fname from module.py
                    py_module = imp.load_source(module_dir_name, mod_file)
                else:
                    py_module = imp.load_compiled(module_dir_name, mod_file)
                
                if module_info:
                    py_module.properties = module_info
                # Look if it's a valid module
                if not module_info:
                    if hasattr(py_module, u'properties'):
                        module_info = py_module.properties
                    else:
                        _logger.error(u'Missing properties dict in your module file %s. The module won\'t be loaded.' % mod_file)
                        failed_modules_dir.append(module_dir_name)
                        continue
                py_module.properties = module_info
                if u'type' not in module_info and u'module_types_by_daemons' not in module_info:
                    self.logger_module_code_loading.warning(u'[directory=%-20s] Bad module file for %s : missing type entry in properties dict' % (module_dir_name, mod_file))
                    failed_modules_dir.append(module_dir_name)
                    continue
                
                # Look at sys.modules and look which modules are now loaded when they was not before:
                # WARNING: there will be duplicate: entries that match the same lib object
                sys_modules_names_added = self._get_only_new_sys_modules(sys_modules_names_before, sys_modules_code_before)
                
                imported_libs_log = ''
                if sys_modules_names_added:
                    imported_libs_log = 'Did import %s python new libraries (%s).' % (len(sys_modules_names_added), ','.join(sys_modules_names_added))
                self.logger_module_code_loading.debug('[directory=%-20s] [%.3fs] Module code was loaded. %s' % (module_dir_name, time.time() - module_start, imported_libs_log))
                
                self.imported_modules.append(py_module)
                if u'module_types_by_daemons' in module_info:
                    available_module_types = [py_module for py_module in module_info[u'module_types_by_daemons'] if py_module[u'daemon'] == self.modules_type]
                    for module_type_name in available_module_types:
                        mod_type = uniform_module_type(module_type_name.get(u'module_type', u''))
                        available_modules_names.append(mod_type)
                elif u'type' in module_info:
                    mod_type = uniform_module_type(module_info[u'type'])
                    available_modules_names.append(mod_type)
            
            except ImportError as exp:
                if not os.path.exists(init_path_file):
                    _logger.error(u'Failed to import the directory %s as a python module. The module won\'t be loaded.' % module_dir)
                    _logger.error(u' - because of missing file : %s.' % init_path_file)
                    _logger.error(u" - The line 'import %s.my_file'. The module won't work in your module.py file." % module_dir_name)
                else:
                    _logger.error(u'Import module [%s] failed: %s.' % (module_dir_name, exp))
            
            except Exception as exp:
                # Oops, something went wrong here... added in the daemon configuration.
                _logger.error(u'Import module [%s] failed: %s.' % (module_dir_name, exp))
                failed_modules_dir.append(module_dir_name)
        if len(self.imported_modules):
            self.logger_module_code_loading.info('A total of %s Shinken Enterprise modules are available for this daemon/module ( %s ): %s ( on a total of %s, loaded in %.3fs )' % (
                len(self.imported_modules), self.modules_type, ', '.join(available_modules_names), len(modules_dirs), time.time() - before))
    
    
    # For a specific module definition, we want to get the pymodule that match us
    def __get_pymodule_from_mod_conf(self, mod_conf):
        # type: (ModuleConfiguration) -> Optional[ModuleCode]
        if not hasattr(mod_conf, 'module_type'):
            return None
        py_module = self.get_module_code_with_module_type(mod_conf.module_type)
        if py_module:
            return py_module
        _error = u'The module %s of type %s is not available for the daemon/module %s.' % (mod_conf.get_name(), mod_conf.module_type, self.modules_type)
        self.errors.append(_error)
        self.logger_module_instance_creation.warning(_error)
        return None
    
    
    def get_module_code_with_module_type(self, module_type):
        # type: (unicode) -> Optional[ModuleCode]
        module_type = uniform_module_type(module_type)
        
        for py_module in self.imported_modules:
            if py_module.properties.get(u'module_types_by_daemons', None):
                for module_info in [m for m in py_module.properties[u'module_types_by_daemons'] if m.get(u'daemon') == self.modules_type]:
                    if module_type == uniform_module_type(module_info.get(u'module_type', u'unknown module type')):
                        return py_module
            elif u'type' in py_module.properties and uniform_module_type(py_module.properties[u'type']) == module_type:
                return py_module
        
        return None
    
    
    # Try to "init" the given module instance.
    # If late_start, don't look for last_init_try
    # If force_start, don't look at restart timing, and just try to start
    # Returns: state of the module
    def _try_instance_init(self, inst, late_start=False, force_start=False, start_logger=None):
        # type: (Union[BaseModule, WithWorkersAndInventoryModule], bool, bool, Optional[PartLogger]) -> int
        # Someone ask us to force a restart of the module, without looking at timing
        if force_start:
            state = INSTANCE_INIT_STATE.INIT_READY
        else:  # Look if we didn't already try to init since not long
            state = inst.is_ready_to_init(self.logger_module_instance_creation, late_start)
        if state == INSTANCE_INIT_STATE.INIT_READY:
            if start_logger:
                start_logger.info(u'Start to create %s module' % (u'an external' if inst.is_external else (u'a worker based' if inst.is_worker_based else u'an internal')))
            if inst.is_external or inst.is_worker_based:
                inst.create_queues(self.manager_factory)
            is_fail, exp_message = inst.do_init(self.logger_module_instance_creation)
            if is_fail:
                self.did_crash(inst, reason=exp_message, do_log=False)
                return INSTANCE_INIT_STATE.INIT_FAILED
            return INSTANCE_INIT_STATE.INIT_SUCCESSFUL
        else:
            return INSTANCE_INIT_STATE.INIT_SKIPPED
    
    
    # Request to "remove" the given instances list or all if not provided
    def clear_instances(self, insts=None):
        if insts is None:
            insts = self.get_all_instances()  # have to make a copy of the list
        for i in insts:
            self.logger_modules_stop.info(u'Stopping module %s' % i.get_name())
            self.__remove_instance(i.get_name(), self.logger_modules_stop)
    
    
    def clear_instances_by_module_on_type(self, module_on_type):
        modules_list = self.instances.copy()
        for key, module in modules_list.iteritems():
            if module_on_type not in module.get_module_on_type():
                self.instances.pop(key)  # We can pop it's a copy
                self.to_restart.pop(key, None)
    
    
    # A daemon did detect that a module instance did crash and so don't want it anymore.
    # We will need to log it, kill the subprocess if external and set to restart later
    def did_crash(self, inst, reason=u'', logger_with_chapter=None, do_log=True):
        if do_log:
            if logger_with_chapter and reason:
                logger_with_chapter.error(u'%s' % reason)
            elif reason:
                self.logger.error(reason)
        self.set_to_restart(inst, reason=reason)
    
    
    # Put an instance to the restart queue
    # By default if an instance is set to restart, it means it's an error,
    # but in some case it's not (like new modules)
    def set_to_restart(self, inst, reason='', is_an_error=True):
        # type: (BaseModule, unicode, bool) -> None
        already_set_to_restart = inst.get_name() in self.to_restart
        try:
            inst.stop_all()
        except:
            # If external, clean its queues
            if inst.is_external or inst.is_worker_based:
                inst.clear_queues()
                inst.kill_manager()
        if inst.get_internal_state() == ModuleState.FATAL and is_an_error:
            inst.fatal_error_has_been_managed = True
            inst.stop_all()
            return
        self.to_restart[inst.get_name()] = inst
        if is_an_error and (reason or not already_set_to_restart):
            self.__register_module_restart(inst.get_name(), reason)
    
    
    def __get_instance_from_modconf(self, mod_conf):
        mod_name = mod_conf.get_name()
        mod_type = mod_conf.get_type()
        py_module = self.__get_pymodule_from_mod_conf(mod_conf)
        # if we cannot find a suitable code to launch, skip it, we already warn about it
        if not py_module:
            return None
        start = time.time()
        _logger = self.logger_module_instance_creation.get_sub_part(mod_name, register=False).get_sub_part(u'module-type=%s' % mod_type, register=False)
        
        try:
            _logger.debug(u'Start to create the module instance.')
            mod_conf.properties = py_module.properties.copy()
            if u'module_types_by_daemons' in py_module.properties:
                module_type = next((md[u'module_type'] for md in py_module.properties[u'module_types_by_daemons'] if uniform_module_type(md[u'module_type']) == uniform_module_type(mod_type)))
                mod_conf.properties[u'type'] = module_type
            
            argspec = inspect.getargspec(py_module.get_instance)
            if len(argspec.args) > 1:
                params = {}
                if u'module_on' in argspec.args:
                    params[u'module_on'] = self.modules_type
                if u'module_type' in argspec.args:
                    params[u'module_type'] = mod_type
                inst = py_module.get_instance(mod_conf, **params)
            else:
                inst = py_module.get_instance(mod_conf)
            if inst is None:  # None = Bad thing happened :)
                _logger.error(u"The module get_instance() call did not return any instance or does not exist.")
                return None
            assert (isinstance(inst, BaseModule))
            _logger.info(u'[%.3fs] The module is created.' % (time.time() - start))
            return inst
        
        except Exception as exp:
            s = unicode(exp)
            _logger.error(u'[%.3fs] The module creation did fail with an exception: %s. Will try to recreate it in the future.' % (time.time() - start, s))
            
            return None
    
    
    # Actually only arbiter call this method with start_external=False..
    # Create, init and then returns the list of module instances that the caller needs.
    # If an instance can't be created or inited then only log is done.
    # That instance is skipped. The previous modules instance(s), if any, are all cleaned.
    def _get_instances(self):
        with self.instances_lock:
            self.clear_instances()
            
            all_was_start = True
            
            for mod_conf in self.modules:
                inst = self.__get_instance_from_modconf(mod_conf)
                if inst:
                    updated_instances = self.instances.copy()
                    updated_instances[mod_conf.get_name()] = inst
                    self.instances = updated_instances
                else:  # there was an error on this module
                    all_was_start = False
            
            # We should init the modules, but not the:
            # * external ones
            # * with workers based
            # because they can be crashed and so it must be done just before forking
            for inst in self.get_all_instances():
                if not inst.is_external and not inst.is_worker_based:
                    did_init = self._try_instance_init(inst)
                    if did_init == INSTANCE_INIT_STATE.INIT_FAILED:
                        # If the init failed, we put in in the restart queue
                        self.set_to_restart(inst)
                        all_was_start = False
                    elif did_init == INSTANCE_INIT_STATE.INIT_SKIPPED:
                        all_was_start = False
            
            return all_was_start
    
    
    def start_internal_instance(self, inst, late_start=False):
        # But maybe the init failed a bit, so bypass this ones from now
        init_state = self._try_instance_init(inst, late_start=late_start)
        if init_state == INSTANCE_INIT_STATE.INIT_FAILED:
            message = u'The module \'%s\' failed to init' % inst.get_name()
            self.set_to_restart(inst, message)
            return False
        if init_state == INSTANCE_INIT_STATE.INIT_SKIPPED:
            self.set_to_restart(inst, is_an_error=False)
            return False
        return True
    
    
    def start_external_instance(self, inst, late_start=False):
        _logger = self.logger_module_instance_creation.get_sub_part(inst.get_name(), register=False).get_sub_part(u'module-type=%s' % inst.get_module_type(), register=False)
        
        # But maybe the init failed a bit, so bypass this ones from now
        # _logger.debug(u'Start to create the module instance')
        start = time.time()
        did_init = self._try_instance_init(inst, late_start=late_start, start_logger=_logger)
        if did_init == INSTANCE_INIT_STATE.INIT_FAILED:
            self.set_to_restart(inst)
            return
        if did_init == INSTANCE_INIT_STATE.INIT_SKIPPED:
            _logger.debug(u'Creation module instance was skipped')
            self.set_to_restart(inst, is_an_error=False)
            return
        try:
            inst.start(daemon_display_name=self.daemon_display_name)
            _logger.info(u'[%.3fs] The module is created. ' % (time.time() - start))
        except Exception as exp:
            _logger.error(u'[%.3fs] The module creation did fail with an exception: %s. Will try to recreate it in the future.' % (time.time() - start, exp))
            self.set_to_restart(inst)
    
    
    # Launch external instances that are load correctly
    def start_external_instances(self, late_start=False):
        for inst in [inst for inst in self.get_all_instances() if inst.is_external]:
            self.start_external_instance(inst, late_start=late_start)
    
    
    def start_worker_based_instance(self, inst, late_start=False):
        # type: (Union[BaseModule, WithWorkersAndInventoryModule], bool) -> bool
        _logger = self.logger_module_instance_creation.get_sub_part(inst.get_name(), register=False).get_sub_part(u'module-type=%s' % inst.get_module_type(), register=False)
        # _logger.info(u'Starting a worker base module.')
        # But maybe the init failed a bit, so bypass this ones from now
        did_init = self._try_instance_init(inst, late_start=late_start, start_logger=_logger)
        if did_init == INSTANCE_INIT_STATE.INIT_SKIPPED:
            self.set_to_restart(inst, is_an_error=False)
            return False
        if did_init == INSTANCE_INIT_STATE.INIT_FAILED:
            _logger.error(u'The worker base module failed to init.')
            self.set_to_restart(inst)
            return False
        
        # ok, init succeed
        _logger.info(u'The worker base module has started.')
        try:
            inst.start_workers(daemon_display_name=self.daemon_display_name)
            return True
        except Exception as e:
            _logger.error(u'The worker base module failed to init and he raised an exception: %s.  The module will be restarted in the future.' % str(e))
            self.did_crash(inst, reason=str(e))
            return False
    
    
    # Launch external instances that are load correctly
    def start_worker_based_instances(self, late_start=False):
        for inst in [inst for inst in self.get_all_instances() if inst.is_worker_based]:
            self.start_worker_based_instance(inst, late_start=late_start)
    
    
    # Request to cleanly remove the given instance.
    # If instance is external also shutdown it cleanly
    def __remove_instance(self, instance_name, update_stop_logger):
        with self.instances_lock:
            inst = self.instances.get(instance_name, None)
            if inst is None:
                _logger = update_stop_logger.get_sub_part(instance_name, register=False)
                _logger.warning(u'Trying to remove the module %s but it is not found in the current running modules: %s' % (instance_name, self.instances))
                return
            
            _logger = update_stop_logger.get_sub_part(inst.get_name(), register=False).get_sub_part(u'module-type=%s' % inst.get_module_type(), register=False)
            try:
                inst.stop_all(_logger)
            except Exception as exp:
                _logger.error(u'The module %s failed to stop and raised an exception : %s' % (instance_name, str(exp)))
                _logger.print_stack()
            # Then do not listen anymore about it
            updated_instances = self.instances.copy()
            del updated_instances[instance_name]
            self.instances = updated_instances
            # SEF-6521: do not forget to clean in all lists:
            if instance_name in self.to_restart:
                del self.to_restart[instance_name]
    
    
    def check_alive_instances(self, skip_external=False):
        # type: (bool) -> None
        # Only for external
        for instance_name, inst in self.get_all_instances_with_name():
            # skip already to restart one
            _logger = self.logger_modules_instance_crash.get_sub_part(inst.get_name(), register=False).get_sub_part(u'module-type=%s' % inst.get_module_type(), register=False)
            
            if instance_name in self.to_restart or inst.fatal_error_has_been_managed:
                continue
            
            # The skip_external is used in a child process because there is no possible to set an external module to a subprocess/submodule
            if skip_external and inst.is_external:
                continue
            
            if not inst.is_alive():
                self.did_crash(inst, u'The module %s goes down unexpectedly!' % instance_name, logger_with_chapter=_logger)
                inst.stop_process()
                continue
            
            if inst.is_worker_based and not inst.check_worker_processes():
                self.did_crash(inst, u'The module %s worker(s) goes down unexpectedly!' % instance_name, logger_with_chapter=_logger)
                continue
            
            # Now look for man queue size. If above value, the module should got a huge problem
            # and so bailout. It's not a perfect solution, more a watchdog
            # If max_queue_size is 0, don't check this
            if self.max_queue_size == 0:
                continue
            # Ok, go launch the dog!
            queue_size = 0
            try:
                queue_size = inst.to_q.qsize()
            except Exception:
                pass
            if queue_size > self.max_queue_size:
                self.did_crash(inst, u'The external module %s got a too high queue size ( %s > %s )!' % (instance_name, queue_size, self.max_queue_size), logger_with_chapter=_logger)
    
    
    def is_instance_set_to_restart(self, instance):
        return instance in self.to_restart.values()
    
    
    # We are looking at the
    def try_to_restart_deads(self, force_start=False):
        to_restart = self.to_restart
        self.to_restart = {}
        for instance_name, inst in to_restart.items():
            if inst.is_external:
                self.start_external_instance(inst)
            elif inst.is_worker_based:
                self.start_worker_based_instance(inst)
            else:
                if not self._try_instance_init(inst, force_start=force_start) == INSTANCE_INIT_STATE.INIT_SUCCESSFUL:
                    self.set_to_restart(inst)
    
    
    def __register_module_restart(self, module_name, reason=u''):
        with self.last_restarts_lock:
            if module_name not in self.last_restarts:
                self.last_restarts[module_name] = []
            
            self.last_restarts[module_name].append({u'timestamp': time.time(), u'reason': reason.rstrip()})
    
    
    # Called from StateCleaner thread
    def clean_modules_states(self):
        with self.last_restarts_lock:
            now_ts = time.time()
            clean_ts = now_ts - self.last_restarts_keep
            
            for last_restarts in self.last_restarts.values():
                delete_count = 0
                
                for restart_ts in last_restarts:
                    if restart_ts < clean_ts:
                        delete_count += 1
                    else:
                        break
                
                del last_restarts[:delete_count]
    
    
    def get_modules_states(self):
        states = []
        for name, inst in self.get_all_instances_with_name():
            start_time = time.time()
            status = {u'restarts': self.last_restarts.get(name, []), u'name': name, u'type': inst.get_module_type()}
            status.update(inst.get_state())
            states.append(status)
            self.logger.log_perf(start_time, u'GET_MODULES_STATES', u'Module %s stats requested' % name, min_time=1, warn_time=1.8)
        return {u'modules': states, u'errors': self.errors}
    
    
    def _get_instances_with_condition(self, condition=lambda inst: True):
        return sorted(filter(condition, self.get_all_instances()), key=lambda inst: self._idx_modules_order_by_name[inst.get_name()])
    
    
    # Do not give to others inst that got problems
    def get_internal_instances(self, phase=None):
        return self._get_instances_with_condition(lambda inst: not inst.is_external and phase in inst.phases and inst not in self.to_restart.values() and inst.get_internal_state() != ModuleState.FATAL)
    
    
    def get_external_instances(self, phase=None):
        return self._get_instances_with_condition(lambda inst: inst.is_external and phase in inst.phases and inst not in self.to_restart.values() and inst.get_internal_state() != ModuleState.FATAL)
    
    
    def get_external_to_queues(self):
        return [inst.to_q for inst in self.get_all_instances() if inst.is_external and inst not in self.to_restart.values() and inst.get_internal_state() != ModuleState.FATAL]
    
    
    def get_external_modules_and_queues(self):
        return [(inst, inst.to_q) for inst in self.get_all_instances() if inst.is_external and inst not in self.to_restart.values() and inst.get_internal_state() != ModuleState.FATAL]
    
    
    def get_external_modules_and_from_queues(self):
        return [(inst, inst.from_module_to_main_daemon_queue) for inst in self.get_all_alive_instances() if (inst.is_external or inst.is_worker_based)]
    
    
    def get_external_from_queues(self):
        return [inst.from_module_to_main_daemon_queue for inst in self.get_all_instances() if (inst.is_external or inst.is_worker_based) and inst not in self.to_restart.values() and inst.get_internal_state() != ModuleState.FATAL]
    
    
    def get_all_alive_instances(self):
        # type: () -> List[BaseModule]
        with self.instances_lock:
            return self._get_instances_with_condition(lambda inst: inst not in self.to_restart.values() and inst.get_internal_state() != ModuleState.FATAL)
    
    
    def stop_all(self, update_stop_logger=None):
        if not update_stop_logger:
            update_stop_logger = self.logger_modules_stop
        update_stop_logger.info(u'Start to shutdown all modules.')
        # Ask internal to quit if they can
        for inst in self.get_internal_instances():
            if not inst.is_worker_based:
                inst.stop_all(update_stop_logger)
        
        # Clear/stop all external & worker based instances
        self.clear_instances([inst for inst in self.get_all_instances() if inst.is_external])
        self.clear_instances([inst for inst in self.get_all_instances() if inst.is_worker_based])
        
        update_stop_logger.info(u'Stopping all modules ended.')
    
    
    def get_all_instances(self):
        with self.instances_lock:
            return self.instances.values()
    
    
    def get_all_instances_name(self):
        with self.instances_lock:
            return self.instances.keys()
    
    
    def get_all_instances_with_name(self):
        with self.instances_lock:
            return self.instances.items()
    
    
    def do_after_fork_cleanup(self, after_fork_new_top_instance):
        if hasattr(after_fork_new_top_instance, u'logger'):
            _logger = after_fork_new_top_instance.logger
        else:
            _logger = self.logger
        cleanup_logger = _logger.get_sub_part(AFTER_FORK_CLEANUP_PART_LOGGER)
        cleanup_logger = cleanup_logger.get_sub_part(u'pid:%s' % os.getpid(), register=False)
        
        # We are in a forked process, modules_manager does no more exist here.
        # * Do not use self.instances_lock as
        #   1 - it is useless here
        #   2 - its state might be incorrect
        # * No iteritems() here as we will delete items in for loop
        for name, inst in self.instances.items():
            module_cleanup_logger = cleanup_logger.get_sub_part(name)
            if inst == after_fork_new_top_instance:
                module_cleanup_logger.debug(u'skipping my own instance cleanup')
                continue
            inst_cleanup = getattr(inst, AFTER_FORK_CLEANUP_METHOD, None)
            if callable(inst_cleanup):
                module_cleanup_logger.debug(u'cleanup of module 〖 %s 〗 is starting' % name)
                try:
                    inst_cleanup()
                except Exception as e:
                    module_cleanup_logger.error(
                        u'On linux system, the forking mechanism (process creation) is fast but create a copy of the father process. So we have to release unnecessary resources inherited from father. The cleaning has been performed but we encountered an error:')
                    module_cleanup_logger.error(u'Cleanup of data from module 〖 %s 〗 raised error 〖 %s 〗' % (name, e))
                    module_cleanup_logger.error(u'Some memory may have not been freed. Shinken will still run if enough memory remains available.')
                    module_cleanup_logger.error(u'You can report this message to support in order to optimize Shinken memory consumption')
                    module_cleanup_logger.print_stack()
                module_cleanup_logger.debug(u'cleanup of module 〖 %s 〗 has ended' % name)
            elif inst_cleanup:
                module_cleanup_logger.error(
                    u'On linux system, the forking mechanism (process creation) is fast but create a copy of the father process. So we have to release unnecessary resources inherited from father. The cleaning has been performed but we encountered an error:')
                module_cleanup_logger.error(u'〖 %s 〗 is not callable, can not run cleanup process' % AFTER_FORK_CLEANUP_METHOD)
                module_cleanup_logger.error(u'Some memory may have not been freed. Shinken will still run if enough memory remains available.')
                module_cleanup_logger.error(u'You can report this message to support in order to optimize Shinken memory consumption')
            module_cleanup_logger.debug(u'removing module 〖 %s 〗 from process memory' % name)
            del self.instances[name]
        del self.instances
